# License: MIT
#
# This is a base class for implementing multi-view estimators using
# co-training.

from abc import abstractmethod

from sklearn.base import BaseEstimator


class BaseCoTrainEstimator(BaseEstimator):
    """
    A base class for multiview co-training. Extends sklearn's
    BaseEstimator.

    Parameters
    ----------
    estimator1 : estimator object, default=None
        The estimator object which will be trained on view 1 of the data.

    estimator2 : estimator object, default=None
        The estimator object which will be trained on view 2 of the data.
        Does not need to be of the same type as estimator1.

    random_state : int
        The starting random seed for fit() and class operations, passed to
        numpy.random.seed().

    Attributes
    ----------
    estimator1_ : estimator object
        The estimator used on view 1.

    estimator2_ : estimator object
        The estimator used on view 2.

    random_state : int
        The starting random seed for fit() and class operations, passed to
        numpy.random.seed().
    """

    def __init__(self,
                 estimator1=None,
                 estimator2=None,
                 random_state=None
                 ):
        self.estimator1_ = estimator1
        self.estimator2_ = estimator2
        self.random_state = random_state

    @abstractmethod
    def fit(self, Xs, y):
        """
        A method to fit co-trained estimators to multiview data.

        Parameters
        ----------
        Xs : list of array-likes or numpy.ndarray
            - Xs length: n_views
            - Xs[i] shape: (n_samples, n_features_i)
            A list of the different views of data to train on.

        y : array, shape (n_samples,)
            Targets of the training data. Unlabeled examples should
            have label np.nan.

        Returns
        -------
        self : returns an instance of self
        """

        return self

    @abstractmethod
    def predict(self, Xs):
        """
        A method to predict the class of multiview data.

        Parameters
        ----------
        Xs : list of array-likes or numpy.ndarray
            - Xs length: n_views
            - Xs[i] shape: (n_samples, n_features_i)
            A list of the different views of data to predict.

        Returns
        -------
        y : array-like (n_samples,)
            Predictions for each sample.
        """

        return self

    def fit_predict(self, Xs, y):
        """
        Fit a co-train estimator to the semi-supervised data and
        then predict.

        Parameters
        ----------
        Xs : list of array-likes or numpy.ndarray
            - Xs length: n_views
            - Xs[i] shape: (n_samples, n_features_i)
             A list of the different views of data to fit and
             then predict.
        y : array, shape (n_samples,)
            Targets of the training data. Unlabeled examples should
            have label np.nan.

        Returns
        -------
        y_pred : array-like (n_samples,)
            Predictions for each sample.
        """

        return self.fit(Xs, y).predict(Xs)
